from pathlib import Path
import shutil
from typing import (
    Optional,
    List,
    Literal,
)
from typeguard import typechecked
from loguru import logger
from IPython.display import display
import pandas as pd
from tidydata.table import Table
from tidydata.schema import SourceSchema
from tidydata.schema import SourceSchema, PipeSchema, ExportSchema, TidyDataSchema
import toml


class TidyData(object):
    """TableManager类"""

    def __getitem__(self, key):
        return self.cache[key]

    @typechecked
    def __setitem__(self, key: str, value: Table):
        self.cache[key] = value

    @typechecked
    def __setitem__(self, key: str, value: Table):
        self.cache[key] = value

    def __enter__(self):
        return self

    def __exit__(self, exception_type, exception_value, exception_traceback):
        self.cache.close()
        shutil.rmtree(self.cache_dir)

    @classmethod
    def from_toml(cls, path: Path | str):
        tidydata = toml.load(path).get("tidydata")
        if tidydata is None:
            raise ValueError(f"Missing key 'tidydata' in file: {path}")
        return cls(**tidydata)

    @typechecked
    def __init__(
        self,
        cache_dir: Path | str = "_cache",
        export_dir: str | Path = "_export",
        log_file: bool = False,
        log_level: Literal[
            "TRACE", "DEBUG", "INFO", "SUCCESS", "WARNING", "ERROR", "CRITICAL"
        ] = "INFO",
        log_format: str = "<green>{time:YYYY-MM-DD HH:mm:ss} {level}</green>: {message}",
        source: Optional[Path | str] = None,
        pipe: Optional[Path | str] = None,
        export: Optional[Path | str] = None,
        keep_cache: bool = True,
        **kwargs,
    ):
        tidydata_schema = TidyDataSchema(
            cache_dir=cache_dir,
            export_dir=export_dir,
            log_file=log_file,
            log_format=log_format,
            log_level=log_level,
            source=source,
            pipe=pipe,
            export=export,
        )

        self.cache_dir = tidydata_schema.cache_dir
        self.export_dir = tidydata_schema.export_dir
        self.log_file = tidydata_schema.log_file
        self.cache = tidydata_schema._cache

        if tidydata_schema.source is not None:
            self.source(tidydata_schema.source)
        if tidydata_schema.pipe is not None:
            self.pipe(tidydata_schema.pipe)
        if tidydata_schema.export is not None:
            self.export(tidydata_schema.export)

        if not keep_cache:
            self.remove()

    @property
    def tags(self):
        return [self.cache.get(key, tag=True)[1] for key in self.keys]

    @property
    def keys(self):
        return list(self.cache.iterkeys())

    def values(self):
        return (self.cache[k] for k in self.keys)

    def key_to_tag(self):
        return pd.DataFrame({"key": self.keys, "tag": self.tags})

    def disk_usage(self):
        return self.cache.volume()

    def close(self):
        logger.debug(f"Closing disk cache in {self.cache_dir}...")
        self.cache.close()
        logger.debug(f"Closed disk cache in {self.cache_dir}...")

    def remove(self):
        self.close()
        logger.debug(f"Removing cache directory: {self.cache_dir}...")
        shutil.rmtree(self.cache_dir)
        logger.info(f"Removed TidyData cache from directory: '{self.cache_dir}'.")

    def pipe(self, toml_file: Path | str):
        PipeSchema.from_toml(toml_file).to_cache(self.cache)
        logger.success(f"All pipe actions took from toml file: '{toml_file}'.")

    def source(self, toml_file: Path | str):
        SourceSchema.from_toml(toml_file).to_cache(self.cache)
        logger.success(f"All sources added from toml file: '{toml_file}'.")

    def export(self, toml_file: Path | str):
        ExportSchema.from_toml(toml_file).export_from_cache(
            self.cache, export_dir=self.export_dir
        )
        logger.success(
            f"All data (metadata or statistics) exported from toml file: '{toml_file}'."
        )

    def get_keys(
        self,
        keys: Optional[str | List[str]] = None,
        tags: Optional[str | List[str]] = None,
    ):
        query_condition = None

        if keys is not None:
            query_condition = f"key.isin({[keys] if isinstance(keys, str) else keys})"

        if tags is not None:
            if query_condition is None:
                query_condition = (
                    f"tag.isin({[tags] if isinstance(tags, str) else tags})"
                )
            else:
                query_condition = (
                    query_condition
                    + "&"
                    + f"tag.isin({[tags] if isinstance(tags, str) else tags})"
                )

        if query_condition is not None:
            queried_keys = self.key_to_tag().query(query_condition)["key"].to_list()

        else:
            queried_keys = self.keys

        return queried_keys

    def describe(
        self,
        keys: Optional[List[str] | str] = None,
        tags: Optional[List[str] | str] = None,
        destype: Literal["metafile", "summary"] = "summary",
    ):
        for i, key in enumerate(self.get_keys(keys=keys, tags=tags), start=1):
            print(f"\033[32m [{i}] Describe Table with name '{key}': ")
            display(self.cache[key].describe(destype=destype))



from pathlib import Path
from typing import Optional, List, Dict, Literal, Iterable, Any
import pandas as pd
import pyreadstat
from ulid import ULID
from pydantic import BaseModel
from tidydata.dtype import DtypeStr, DtypeMappings

from tidydata.utils import (
    is_default_index,
    infer_stata_bin,
    cast_mix_dtype,
    log_error,
    convert_dtype_backend,
)
from tidydata.table import Table
from loguru import logger

ulid = ULID()

FileSuffixStr = Literal[
    ".csv",
    ".tsv",
    ".parquet",
    ".pq",
    ".dta",
    ".pickle",
    ".pkl",
    ".feather",
    ".fea",
    ".xlsx",
    ".xls",
]

FILE_SUFFIES = [
    ".csv",
    ".tsv",
    ".parquet",
    ".pq",
    ".dta",
    ".pickle",
    ".pkl",
    ".feather",
    ".fea",
    ".xlsx",
    ".xls",
]


def read_dta(
    path: str | Path,
    columns: Optional[List[str]] = None,
    binary: Optional[str | Path] = None,
    version: Literal["mp", "se", "be"] = "mp",
    engine: Literal["pandas", "stata-bin"] = "pandas",
):
    from_file = Path(path)

    if not from_file.suffix == ".dta":
        logger.error(f"Not a DTA file: {path}")

    if engine == "stata-bin":
        import stata_setup

        stata_setup.config((binary or infer_stata_bin()), edition=version, splash=False)
        from pystata import stata

        temp_csv = Path(f"TempFile_{ulid.generate()}.csv")

        stata.run(
            f"use {' '.join(columns) if columns is not None else ''} using {from_file},clear",
            quietly=True,
        )
        stata.run(f"export delimited using {temp_csv}, quote datafmt", quietly=True)
        df = pd.read_csv(temp_csv, columns=columns)
        temp_csv.unlink()
        return df

    else:
        return pd.read_stata(from_file, columns=columns)


def read_csv(
    path: str | Path,
    columns: Optional[List[str]] = None,
    sep: str = ",",
):
    return pd.read_csv(
        path, usecols=columns, sep=sep, engine="pyarrow", dtype_backend="pyarrow"
    )


def read_tsv(
    path: str | Path,
    columns: Optional[List[str]] = None,
):
    return pd.read_csv(
        path, usecols=columns, sep=r"\t", engine="pyarrow", dtype_backend="pyarrow"
    )


def read_parquet(
    path: str | Path,
    columns: Optional[List[str]] = None,
):
    return pd.read_parquet(
        path, columns=columns, engine="pyarrow", dtype_backend="pyarrow"
    )


def read_feather(
    path: str | Path,
    columns: Optional[List[str]] = None,
):
    return pd.read_feather(path, columns=columns, dtype_backend="pyarrow")


def read_pickle(
    path: str | Path,
    columns: Optional[List[str]] = None,
):
    obj = pd.read_pickle(path)

    if not isinstance(obj, pd.DataFrame) and columns is not None:
        log_error(
            f"Value conflict: 'columns' should be None when not a pandas DataFrame object in pickle file:{path}"
        )

    if isinstance(obj, pd.DataFrame) and columns is not None:
        return obj[columns]
    else:
        return obj


def read_excel(
    path: str | Path,
    columns: Optional[List[str]] = None,
    sheet: int | str = 0,
):
    return pd.read_excel(path, usecols=columns, sheet_name=sheet)


class StataReadOption(BaseModel):
    binary: Optional[Path | str] = None
    version: Literal["mp", "se", "be"] = "mp"
    engine: Literal["stata-bin", "pyreadstat", "pandas"] = "pandas"


class ExcelReadOption(BaseModel):
    sheet: int | str = 0


class CSVReadOption(BaseModel):
    sep: str = ","


class PickleReadOption(BaseModel):
    validate_table: bool = False


class ParquetReadOption(BaseModel):
    pass


class FeatherReadOption(BaseModel):
    pass


SUFFIX_TO_READERS = {
    ".csv": (read_csv, CSVReadOption),
    ".tsv": (read_tsv, CSVReadOption),
    ".dta": (read_dta, StataReadOption),
    ".xlsx": (read_excel, ExcelReadOption),
    ".xls": (read_excel, ExcelReadOption),
    ".pickle": (read_pickle, PickleReadOption),
    ".pkl": (read_pickle, PickleReadOption),
    ".parquet": (read_parquet, ParquetReadOption),
    ".pq": (read_parquet, ParquetReadOption),
    ".feather": (read_feather, FeatherReadOption),
    ".fea": (read_feather, FeatherReadOption),
}


def read_file(path: str | Path, columns: Optional[List[str]] = None, **options):
    read_func, read_option = SUFFIX_TO_READERS.get(Path(path).suffix, (None, None))
    if read_func is None:
        log_error(
            f"Unsupported File Type: file's suffix in '{path}' is not in {list(SUFFIX_TO_READERS.keys())}"
        )

    return read_func(path, columns=columns, **(read_option(**options).model_dump()))


# SUFFIX_TO_READERS = {
#     ".csv": read_csv,
#     ".dta": read_dta,
#     ".parquet": read_parquet,
#     ".pq": read_parquet,
#     ".tsv": read_tsv,
#     ".feather": read_feather,
#     ".fea": read_feather,
#     ".pickle": read_pickle,
#     ".pkl": read_pickle,
#     ".xlsx": read_excel,
#     ".xls": read_excel,
# }

# SUFFIX_TO_OPTIONS = {
#     ".csv": "csv",
#     ".dta": "stata",
#     ".parquet": "parquet",
#     ".pq": "parquet",
#     ".tsv": "tsv",
#     ".feather": "feather",
#     ".fea": "feather",
#     ".pickle": "pickle",
#     ".pkl": "pickle",
#     ".xlsx": "excel",
#     ".xls": "excel",
# }


class StataWriteOption(BaseModel):
    index: bool | Literal["auto"] = "auto"
    var_labels: Optional[Dict[str,str]] = None


class CSVWriteOption(BaseModel):
    index: bool | Literal["auto"] = "auto"
    sep: str = ","


class TSVWriteOption(BaseModel):
    index: bool | Literal["auto"] = "auto"
    sep: str = "\t"


class ParquetWriteOption(BaseModel):
    index: bool | Literal["auto"] = "auto"
    engine: Literal["auto", "pyarrow", "fastparquet"] = "auto"


# Start Here
class ExcelWriteOption(BaseModel):
    index: bool | Literal["auto"] = "auto"
    sheet: str = "Sheet1"
    engine: Literal["openpyxl", "xlsxwriter"] = None


class FeatherWriteOption(BaseModel):
    index: bool | Literal["auto"] = "auto"


def write_parquet(
    data: pd.DataFrame | Table,
    to_file: Path | str,
    index: bool | Literal["auto"] = "auto",
    engine: Literal["auto", "pyarrow", "fastparquet"] = "auto",
):
    if index == "auto":
        index = False if is_default_index(data.index) else True
    if isinstance(data, Table):
        data.data.to_parquet(to_file, index=index, engine=engine)
    else:
        data.to_parquet(to_file, index=index, engine=engine)


def write_csv(
    data: pd.DataFrame | Table,
    to_file: Path | str,
    index: bool | Literal["auto"] = "auto",
    sep: str = ",",
):
    if index == "auto":
        index = False if is_default_index(data.index) else True

    if isinstance(data, Table):
        data.data.to_csv(to_file, index=index, sep=sep)
    else:
        data.to_csv(to_file, index=index, sep=sep)


def write_excel(
    data: pd.DataFrame | Table,
    to_file: Path | str,
    index: bool | Literal["auto"] = "auto",
    sheet: str = "Sheet1",
    engine: Literal["openpyxl", "xlsxwriter"] = None,
):
    if index == "auto":
        index = False if is_default_index(data.index) else True

    if isinstance(data, Table):
        data.data.to_excel(to_file, index=index, sheet_name=sheet, engine=engine)
    else:
        data.to_excel(to_file, index=index, sheet_name=sheet, engine=engine)


def write_feather(
    data: pd.DataFrame | Table,
    to_file: Path | str,
    index: bool | Literal["auto"] = "auto",
):
    if index == "auto":
        index = False if is_default_index(data.index) else True

    if isinstance(data, Table):
        data.data.to_feather(to_file, index=index)
    else:
        data.to_feather(to_file, index=index)


def write_dta(
    data: pd.DataFrame | Table,
    to_file: Path | str,
    index: bool | Literal["auto"] = "auto",
    var_labels: Optional[Dict[str,str]] = None,
):
    if index == "auto":
        index = False if is_default_index(data.index) else True

    if isinstance(data, Table):
        df = data.data
        data_label = data.description[:80]
        column_labels = data.column_labels.to_dict()
        if var_labels is not None:
            column_labels.update(var_labels)
  
    else:
        df = data
        data_label = None
        column_labels = var_labels

    try:
        df.to_stata(
            to_file,
            data_label=data_label,
            version=118,
            write_index=index,
            variable_labels=column_labels
        )
    except:
        convert_dtype_backend(df, backend="numpy").to_stata(
            to_file,
            data_label=data_label,
            version=118,
            write_index=index,
            variable_labels=column_labels,
        )


SUFFIX_TO_WRITERS = {
    ".parquet": (write_parquet, ParquetWriteOption),
    ".pq": (write_parquet, ParquetWriteOption),
    ".csv": (write_csv, CSVWriteOption),
    ".tsv": (write_csv, TSVWriteOption),
    ".dta": (write_dta, StataWriteOption),
    ".xlsx": (write_excel, ExcelWriteOption),
    ".xls": (write_excel, ExcelWriteOption),
    ".feather": (write_feather, FeatherWriteOption),
    ".fea": (write_feather, FeatherWriteOption),
}


def write_file(data: pd.DataFrame | Table, to_file: str | Path, **options):
    write_func, option_schema = SUFFIX_TO_WRITERS.get(Path(to_file).suffix)

    if write_func is None:
        logger.error(
            f"Unsupported File Type: file's suffix in '{to_file}' is not in {list(SUFFIX_TO_WRITERS.keys())}"
        )

    option = option_schema(**options).model_dump()

    return write_func(data=data, to_file=to_file, **option)


def scan_dir(directory: Path | str, suffies: Optional[Iterable[FileSuffixStr]] = None):
    source_dir = Path(directory)
    if not source_dir.is_dir():
        logger.error(f"Invalid directory: {directory}")

    if suffies is not None:
        return (f for suffix in suffies for f in source_dir.rglob(f"*{suffix}"))
    else:
        return (f for f in source_dir.rglob("*"))


class FileConverter(object):
    """文件转换器"""

    def __init__(
        self,
        source: Path | str | Dict[str, Path | str],
        suffies: List[FileSuffixStr] = None,
        to_dir: str | Path = Path.cwd(),
        quietly: bool = False,
        **options,
    ) -> None:
        self.suffies = FILE_SUFFIES if suffies is None else suffies
        self.source = source
        self.to_dir = Path(to_dir)
        self.options = {} if options is None else options
        self.quietly = quietly

        if isinstance(self.source, dict):
            for key, path in source.items():
                if not Path(path).is_file() or Path(path).suffix not in set(
                    self.suffies
                ):
                    logger.error(
                        f"Invalid file path in key '{key}': {path} does not exists or not with suffies in {self.suffies}"
                    )

        else:
            self.source = {
                f.stem: f
                for suffix in self.suffies
                for f in Path(self.source).rglob(f"*{suffix}")
            }

        if self.to_dir.exists() and not self.to_dir.is_dir():
            logger.error(f"Not a directory in 'to_dir': {to_dir}")

        Path(self.to_dir).mkdir(exist_ok=True)

        if not self.quietly:
            logger.info(
                f"==> Created a FileConverter from source {self.source} to Directory {self.to_dir}"
            )

    def to_parquet(
        self,
        keys: Optional[Dict[str, str] | List[str]] = None,
        to_dir: Optional[Path | str] = None,
        quietly: bool = False,
    ):
        keys = list(self.source.keys()) if keys is None else keys
        parquet_mappings = dict(zip(keys, keys)) if isinstance(keys, list) else keys

        to_dir = Path(to_dir) if to_dir is not None else self.to_dir
        Path(to_dir).mkdir(exist_ok=True)
        for from_key, to_key in parquet_mappings.items():
            to_file = to_dir / f"{to_key}.parquet"
            read_file(self.source[from_key], **self.options).to_parquet(to_file)

            if not quietly:
                logger.info(
                    f"==> Converted '{from_key}' to Parquet file {to_file} in Directory {to_dir}"
                )



from pydantic import (
    BaseModel,
    model_validator,
    AnyHttpUrl,
)
from typing import (
    Optional,
    Dict,
    List,
    Hashable,
    Tuple,
    Any,
    Literal,
    Callable,
)
from tidydata.dtype import DtypeStr
from pathlib import Path
from tidydata.io import read_file
from loguru import logger
from ast import literal_eval
import toml
import pandas as pd
from tidydata.table import Table
from tidydata.io import SUFFIX_TO_WRITERS
from diskcache import Cache
from tidydata.utils import log_error, get_duplicate_items
from diskcache import Cache
import shutil
import sys


class ColumnMeta(BaseModel):
    columns: Optional[List[str]] = None
    column_labels: Optional[Dict[str, str]] = None
    column_dtypes: Optional[Dict[str, DtypeStr]] = None
    column_aliases: Optional[Dict[str, str]] = None
    value_ranges: Optional[Dict[str, Tuple | List | Dict[int, str]]] = None
    value_repls: Optional[Dict[str, Dict[Hashable, Any]]] = None
    value_nafills: Optional[Dict[str, Any]] = None

    @classmethod
    def from_file(cls, path: Path | str):
        df = (
            pd.read_csv(
                path,
                dtype={
                    "Is_Use": "bool[pyarrow]",
                    "Column": "string[pyarrow]",
                    "Column_Label": "string[pyarrow]",
                    "Column_Dtype": "string[pyarrow]",
                },
            )
            .dropna(subset=["Column"])
            .query("Is_Use==True")
        )

        columns = df["Column"]

        if not columns.is_unique:
            log_error(f"Not unique values in 'Column' from file {path}")
        df.set_index("Column", inplace=True, drop=True)
        columns = columns.to_list()

        # validate column labels
        column_labels = df["Column_Label"]

        if not column_labels.is_unique:
            log_error(f"Not unique values in 'Column_Label' column from file {path}")
        if column_labels.hasnans:
            log_error(f"NA value in 'Column_Label; from file {path}")
        column_labels = column_labels.to_dict()

        # validate column dtypes
        column_dtypes = df["Column_Dtype"]

        if column_dtypes.hasnans:
            logger.error(f"NA value in 'Column_Dtype; from file {path}")
        column_dtypes = column_dtypes.to_dict()

        # validate column aliases
        column_aliases = df.get("Column_Alias")
        if isinstance(column_aliases, pd.Series):
            column_aliases = column_aliases.dropna()
            column_aliases = (
                column_aliases.to_dict() if not column_aliases.empty else None
            )
        # validate value ranges
        value_ranges = df.get("Value_Range")
        if isinstance(value_ranges, pd.Series):
            value_ranges = value_ranges.dropna()
            value_ranges = (
                value_ranges.map(eval).to_dict() if not value_ranges.empty else None
            )

        # validate value prerepls
        value_repls = df.get("Value_PreRepl")
        if isinstance(value_repls, pd.Series):
            value_repls = value_repls.dropna()
            value_repls = (
                value_repls.map(literal_eval).to_dict()
                if not value_repls.empty
                else None
            )

        # validate value prefill
        value_prefills = df.get("Value_PreFill")
        if isinstance(value_prefills, pd.Series):
            value_prefills = value_prefills.dropna()
            value_prefills = (
                value_prefills.to_dict() if not value_prefills.empty else None
            )

        return ColumnMeta(
            columns=columns,
            column_labels=column_labels,
            column_dtypes=column_dtypes,
            column_aliases=column_aliases,
            value_ranges=value_ranges,
            value_repls=value_repls,
            value_nafills=value_prefills,
        )


class SourceData(BaseModel):
    name: str
    description: str
    tag: Optional[str] = None
    datafile: str | Path
    metafile: Optional[str | Path] = None
    docs: Optional[Path | str | AnyHttpUrl] = None
    reader: Dict = {}
    replace: bool = True
    _colmeta: Optional[ColumnMeta] = None

    @model_validator(mode="after")
    def load_colmeta(self):
        self._colmeta = (
            ColumnMeta.from_file(self.metafile)
            if self.metafile is not None
            else ColumnMeta()
        )
        return self

    @model_validator(mode="after")
    def update_tag(self):
        if self.tag is None:
            self.tag = self.name
        return self

    @classmethod
    def from_datafile(cls, source: Path | str):
        return cls(
            name=source.stem,
            datafile=source,
            description=f"Auto-generated table '{source.stem}' from file: {source}",
        )

    def to_cache(self, cache: Cache):
        if not self.replace:
            cached_keys = set(cache.iterkeys())
            if self.name in cached_keys:
                log_error(
                    f"Value in 'name' should not be in cached_keys when 'replace' is False: {self.name}."
                )

        logger.debug(
            f"==> Constructing Table object '{self.name}' with description: {self.description}."
        )

        cache.set(
            key=self.name,
            value=Table.from_file(
                path=self.datafile,
                name=self.name,
                tag=self.tag,
                description=self.description,
                columns=self._colmeta.columns,
                column_labels=self._colmeta.column_labels,
                column_dtypes=self._colmeta.column_dtypes,
                column_aliases=self._colmeta.column_aliases,
                value_repls=self._colmeta.value_repls,
                value_nafills=self._colmeta.value_nafills,
                value_ranges=self._colmeta.value_ranges,
                **self.reader,
            ),
            tag=self.tag,
        )

        logger.debug(
            f"==> Added Table with key '{self.name}' and tag '{self.tag}' to cache {cache.directory}"
        )


class SourceSchema(BaseModel):
    data: List[Dict[str, Any]]

    @model_validator(mode="after")
    def parse_data(self):
        self.data = [SourceData(**dt) for dt in self.data]
        return self

    @model_validator(mode="after")
    def check_duplicate_keys(self):
        logger.debug(f"==> Validating the uniqueness of table names...")
        duplicate_keys = get_duplicate_items((src.name for src in self.data))
        if len(duplicate_keys) > 0:
            log_error(f"Duplicate keys {duplicate_keys} exists in SourceSchema's data")
        logger.debug(f"==> Successfully constructed a SourceSchema object.")
        return self

    @classmethod
    def from_toml(cls, source: Path | str):
        source_path = Path(source)

        if not source_path.is_file() or not source_path.suffix == ".toml":
            log_error(f"Not an existing toml file: {source}")

        logger.debug(f"==>Constructing SourceSchema from toml file: {source}")
        logger.debug(f"==> Getting the key 'source'...")
        source_dict = toml.load(source_path).get("source")
        if source_dict is None:
            log_error(f"Missing key 'source' in toml file: {source}")

        logger.debug(f"==> Getting the the nested key 'data' in 'source'...")
        source_data = source_dict.get("data")
        if source_data is None:
            log_error(f"Missing nest Key 'data' in 'source': {source_data}")
        logger.debug(
            f"==> Checking whether the value of key 'data' in 'source' is a list..."
        )
        if not isinstance(source_data, list):
            log_error(f"Not a List value for 'data': {source_data}")

        return SourceSchema(data=source_data)

    def to_cache(self, cache: Cache):
        for dt in self.data:
            dt.to_cache(cache)
            logger.info(f"Added source: {dt.description} as key '{dt.name}'.")


class ReshapeArgs(BaseModel):
    to_shape: Literal["long", "wide"]
    groups: str | List[str] | Dict[str, List[str]]
    i: Optional[str | List[str]] = None
    j: Optional[str | List[str]] = None
    sep: str = r"@"
    suffix: str = r"\d+"
    dropna: bool = False


class MutateArgs(BaseModel):
    exprs: List[str]
    repl_dict: Optional[Dict[str, str]] = None
    engine: Optional[Literal["numexpr", "python"]] = None


class FormatArgs(BaseModel):
    description: Optional[str] = None
    columns: Optional[List[str] | str] = None
    rows: Optional[Any] = None
    column_labels: Optional[Dict[str, str]] = None
    column_dtypes: Optional[Dict[str, DtypeStr]] = None
    value_repls: Optional[Dict[str, Dict[Hashable, Any]]] = None
    value_nafills: Optional[Dict[str, Any]] = None
    value_ranges: Optional[Dict[str, Tuple | List | Dict[int, str]]] = None
    dtype_backend: Optional[Literal["numpy", "numpy_nullable", "pyarrow"]] = None
    drop_duplicates: Optional[List[str] | Literal["all"]] = None
    column_aliases: Optional[Dict[str, str]] = None


class ConcatArgs(BaseModel):
    others: List[str]
    by_axis: Literal["column", "row"]
    on: Optional[List[str]] = None
    how: Literal["left", "inner", "outer"] = "outer"
    keycol: Optional[Dict[str, List[int | str]]] = None


PIPE_ARG_SCHEMA = {
    "reshape": ReshapeArgs,
    "concat": ConcatArgs,
    "mutate": MutateArgs,
    "format": FormatArgs,
}


class PipeAction(BaseModel):
    description: str
    from_keys: str | List[str]
    to_keys: Optional[str | List[str]] = None
    to_tags: Optional[str | List[str]] = None
    func: Literal["reshape", "concat", "mutate", "format"]
    func_args: Dict = {}
    split_mode: bool = False
    replace: bool = True
    _method: Optional[Callable] = None

    @model_validator(mode="after")
    def validate_keys(self):
        logger.debug(f"Validating pipe action: {self.description} ")
        logger.debug(
            f"==> Validating whether the length of 'from_keys' match with the length of 'to_keys'..."
        )

        if isinstance(self.from_keys, str):
            self.from_keys = [self.from_keys]

        if self.to_keys is None:
            self.to_keys = self.from_keys

        if isinstance(self.to_keys, str):
            self.to_keys = [self.to_keys]

        if len(self.from_keys) != len(self.to_keys):
            log_error(
                f"Length mismatch: 'from_keys' and 'to_keys' should have the same length in a pipe action."
            )

        return self

    @model_validator(mode="after")
    def parse_tags(self):
        logger.debug(f"==> Parsing tags from 'to_tags'...")
        if self.to_tags is None:
            self.to_tags = self.to_keys
        else:
            if isinstance(self.to_tags, str):
                self.to_tags = [self.to_tags] * len(self.to_keys)

            if len(self.to_keys) != len(self.to_tags):
                log_error(
                    f"Length mismatch: 'to_tags' and 'to_keys' should have the same length in a pipe action."
                )
        return self

    @model_validator(mode="after")
    def parse_func_args(self):
        func_schema = PIPE_ARG_SCHEMA[self.func]

        if self.split_mode:
            logger.debug(
                f"==> Parsing 'func_args' as a dict with splited function '{self.func}' schemas..."
            )

            if len(self.func_args) > 0:
                extra_func_arg_keys = set(self.func_args.keys()) - set(self.from_keys)
                if len(extra_func_arg_keys) > 0:
                    log_error(
                        f"Keys in the dict 'func_args' are out of the scope of 'from_keys': {extra_func_arg_keys}."
                    )

            self.func_args = {
                k: func_schema(**v).model_dump() for k, v in self.func_args.items()
            }

        else:
            logger.debug(
                f"==> Parsing 'func_args' as a dict with unified function '{self.func}' schema..."
            )
            self.func_args = {
                k: func_schema(**self.func_args).model_dump() for k in self.from_keys
            }

        logger.debug(
            f"==>Successfully parsed 'func_args' as a dict of function schema: {self.func_args}"
        )

        return self

    @model_validator(mode="after")
    def parse_method(self):
        logger.debug(f"==> Getting method Table from 'func': {self.func}")
        self._method = getattr(Table, self.func)
        return self

    def to_cache(self, cache: Cache):
        cached_keys = set(cache.iterkeys())
        extra_from_keys = set(self.from_keys) - cached_keys

        if len(extra_from_keys) > 0:
            log_error(
                f"values in 'from_keys' are not in cached keys: {extra_from_keys}"
            )

        if not self.replace:
            duplicated_to_keys = set(self.to_keys).intersection(cached_keys)
            if len(duplicated_to_keys) > 0:
                log_error(
                    f"values in 'to_keys' are in cached keys: {duplicated_to_keys}"
                )

        for from_key, to_key, to_tag in zip(self.from_keys, self.to_keys, self.to_tags):
            if self.func == "concat":
                logger.debug(
                    f"==> Updating 'others' in 'func_args' with a Table generator related to its keys..."
                )
                self.func_args[from_key]["others"] = (
                    cache[other_key] for other_key in self.func_args[from_key]["others"]
                )

            logger.debug(
                f"==> Applying pipe action from key '{from_key}' to '{to_key}'..."
            )

            cache.set(
                key=to_key,
                value=self._method(cache[from_key], **self.func_args.get(from_key, {}))
                .reset_table_name(to_key)
                .reset_table_tag(to_tag),
                tag=to_tag,
            )

            logger.debug(f"==> Added piped Table '{to_key}' to cache {cache.directory}")


class PipeSchema(BaseModel):
    actions: List[Dict[str, Any]]

    @model_validator(mode="after")
    def generate_actions(self):
        self.actions = [PipeAction(**ac) for ac in self.actions]
        return self

    @classmethod
    def from_toml(cls, pipe: Path | str):
        pipe_path = Path(pipe)
        if not pipe_path.is_file() or not pipe_path.suffix == ".toml":
            log_error(f"Not an existing toml file: {pipe}")

        logger.debug(f"Constructing PipeSchema from toml file: {pipe}")
        logger.debug(f"==> Getting the key 'pipe'...")
        pipe_dict = toml.load(pipe_path).get("pipe")
        if pipe_dict is None:
            log_error(f"Missing key 'pipe' in toml file: {pipe}")

        logger.debug(f"==> Getting the the nested key 'action' in 'pipe'...")
        pipe_actions = pipe_dict.get("action")
        if pipe_actions is None:
            log_error(f"Missing nest Key 'action' in 'pipe' from file: {pipe}")
        logger.debug(
            f"==> Checking whether the value of key 'action' in 'pipe' is a list..."
        )
        if not isinstance(pipe_actions, list):
            log_error(f"Not a List value for 'action' in 'pipe': {pipe_actions}")

        return PipeSchema(actions=pipe_actions)

    def to_cache(self, cache: Cache):
        for action in self.actions:
            action.to_cache(cache)
            logger.info(
                f"Took pipe action: {action.description} as keys {action.to_keys}."
            )


class ExportData(BaseModel):
    name: str
    description: str
    datafile: bool | Path | str = False
    metafile: bool | Path | str = False
    statfile: bool | Path | str = False
    rows: Optional[str | int | List[str]] = None
    columns: Optional[str | List[str]] = None
    by_columns: Optional[str | List[str]] = None
    by_rows: Optional[int] = None
    writer: Dict[str, Any] = {}
    replace: bool = True

    @model_validator(mode="after")
    def check_file_path(self):
        # Ensure at least one file path
        if (self.datafile, self.metafile, self.statfile) == (False, False, False):
            log_error(
                f"Values of 'datafile', 'metafile' and 'statfile' should not be all False."
            )

        if (self.by_columns or self.by_rows) is not None:
            # Ensure file path is an existing file path when spliting data by columns or by rows
            if isinstance(self.datafile, (str, Path)) and Path(self.datafile).is_file():
                log_error(
                    f"Value in 'datafile' should not be a file path. 'datafile' should be a directory or bool value when 'by_columns' and 'by_rows' are not all None."
                )
            if isinstance(self.metafile, (str, Path)) and Path(self.metafile).is_file():
                log_error(
                    f"Value in 'metafile' should not be a file path. 'metafile' should be a directory or bool value when 'by_columns' and 'by_rows' are not all None."
                )
            if isinstance(self.statfile, (str, Path)) and Path(self.statfile).is_file():
                log_error(
                    f"Value in 'statfile' should not be a file path. 'statfile' should be a directory or bool value when 'by_columns' and 'by_rows' are not all None."
                )

        return self

    def export_from_cache(self, cache: Cache, export_dir: Path | str):
        logger.debug(f"==> Checking main export path: '{export_dir}'...")
        export_dir = Path(export_dir)
        export_dir.mkdir(exist_ok=True)

        logger.debug(f"==> Checking values of 'datafile', 'metafile' and 'statfile'...")

        if self.datafile is True:
            logger.debug(f"==> Setting default datafile path...")
            self.datafile = export_dir / f"datafile-{self.name}.csv"
            logger.debug(f"==> Set default datafile path as '{self.datafile}'.")
        if self.metafile is True:
            logger.debug(f"==> Setting default metafile path...")
            self.metafile = export_dir / f"metafile-{self.name}.csv"
            logger.debug(f"==> Set default metafile path as '{self.metafile}'.")
        if self.statfile is True:
            logger.debug(f"==> Setting default statfile path...")
            self.statfile = export_dir / f"statfile-{self.name}.csv"
            logger.debug(f"==> Set default statfile path as '{self.statfile}'.")

        if not self.replace:
            logger.debug(
                f"==> Checking the conflict whether export file exists when replace is False..."
            )
            if isinstance(self.datafile, Path) and self.datafile.exists():
                log_error(
                    f"Path exists: {self.datafile}. Path of 'datafile' should not exists when 'replace' is False."
                )

            if isinstance(self.metafile, Path) and self.metafile.exists():
                log_error(
                    f"Path exists: {self.metafile}. Path of 'metafile' should not exists when 'replace' is False."
                )
            if isinstance(self.statfile, Path) and self.statfile.exists():
                log_error(
                    f"Path exists: {self.statfile}. Path of 'statfile' should not exists when 'replace' is False."
                )

        logger.debug(
            f"==> Getting Table '{self.name}' from cache and filtering by conditions of columns and rows..."
        )

        table = cache[self.name].select_columns(self.columns).select_rows(self.rows)

        if (self.by_columns or self.by_rows) is not None:
            logger.debug(f"==> Spliting Table '{self.name}' and exporting...")
            for k, tab in table.split(by_columns=self.by_columns, by_rows=self.by_rows):
                if self.datafile is not False:
                    self.datafile.mkdir(exist_ok=True)
                    to_file = self.datafile / f"{k}.parquet"
                    tab.to_datafile(to_file)
                    logger.info(
                        f"Exported splited datafile: {self.description} from key '{self.name}' to '{to_file}'."
                    )

                if self.metafile is not False:
                    self.metafile.mkdir(exist_ok=True)
                    to_file = self.metafile / f"metafile-{k}.csv"
                    tab.to_metafile(to_file)
                    logger.info(
                        f"Exported splited metafile: {self.description} from key '{self.name}' to '{to_file}'."
                    )
                if self.statfile is not False:
                    self.statfile.mkdir(exist_ok=True)
                    to_file = self.statfile / f"statfile-{k}.csv"
                    tab.to_statfile(to_file)
                    logger.info(
                        f"Exported splited statfile: {self.description} from key '{self.name}' to '{to_file}'."
                    )
        else:
            if self.datafile is not False:
                table.to_datafile(self.datafile, **self.writer)
                logger.info(
                    f"Exported datafile: {self.description} from key '{self.name}' to '{self.datafile}'."
                )
            if self.metafile is not False:
                table.to_metafile(self.metafile, **self.writer)
                logger.info(
                    f"Exported metafile: {self.description} from key '{self.name}' to '{self.metafile}'."
                )
            if self.statfile is not False:
                table.to_statfile(self.statfile, **self.writer)
                logger.info(
                    f"Exported statfile: {self.description} from key '{self.name}' to '{self.statfile}'."
                )

        del table


class ExportSchema(BaseModel):
    data: List[Dict]

    @model_validator(mode="after")
    def parse_data(self):
        logger.debug(f"==> Generating the list ExportData objects from 'data'...")
        self.data = [ExportData(**dt) for dt in self.data]

        return self

    def check_duplicate_keys(self):
        logger.debug(f"==> Validating the uniqueness of export file names...")
        duplicate_keys = get_duplicate_items((src.name for src in self.data))
        if len(duplicate_keys) > 0:
            log_error(f"Duplicate keys {duplicate_keys} exists in ExportSchema's data")
        logger.debug(f"==> Successfully constructed a ExportSchema object.")
        return self

    @classmethod
    def from_toml(cls, export: Path | str):
        export_path = Path(export)

        if not export_path.is_file() or not export_path.suffix == ".toml":
            log_error(f"Not an existing toml file: {export}")

        logger.debug(f"==>Constructing ExportSchema from toml file: {export}")
        logger.debug(f"==> Getting the key 'export'...")
        export_dict = toml.load(export_path).get("export")
        if export_dict is None:
            log_error(f"Missing key 'export' in toml file: {export}")

        return ExportSchema(**export_dict)

    def export_from_cache(self, cache: Cache, export_dir: Path | str):
        for dt in self.data:
            dt.export_from_cache(cache=cache, export_dir=export_dir)


class TidyDataSchema(BaseModel):
    cache_dir: Path | str = "_cache"
    export_dir: Path | str = "_export"
    log_file: bool = False
    log_level: Literal[
        "TRACE", "DEBUG", "INFO", "SUCCESS", "WARNING", "ERROR", "CRITICAL"
    ] = "INFO"
    log_format: str = "<green>{time:YYYY-MM-DD HH:mm:ss} {level}</green>: {message}"
    source: Optional[Path | str] = None
    pipe: Optional[Path | str] = None
    export: Optional[Path | str] = None
    keep_cache: bool = True
    _cache: Optional[Cache] = None

    @model_validator(mode="after")
    def parse_dirs(self):
        logger.remove()
        logger.add(
            sys.stdout,
            format=self.log_format,
            level=self.log_level,
            colorize=True,
        )

        self.export_dir = Path(self.export_dir)
        self.export_dir.mkdir(exist_ok=True)
        self.cache_dir = Path(self.cache_dir)

        if self.log_file:
            self.log_file = self.export_dir / f"tidydata-{self.cache_dir.stem}.log"
            logger.add(
                self.log_file,
                format=self.log_format,
                level=self.log_level,
            )

        logger.info(f"Initializing TidyData cache from directory '{self.cache_dir}'...")

        logger.debug(f"==> Validating and connecting cache {self.cache_dir}....")

        if self.cache_dir.exists():
            self._cache = Cache(
                directory=self.cache_dir,
                size_limit=int(shutil.disk_usage(".").free * 0.95),
                cull_limit=0,
                tag_index=True,
            )

            broken_messages = self._cache.check()
            if len(broken_messages) > 0:
                for m in broken_messages:
                    log_error(m.message.__str__())

            for k in self._cache.iterkeys():
                if not isinstance(self._cache[k], Table):
                    log_error(
                        f"Key {k} is not a Table object in cache {self.cache_dir}"
                    )

        else:
            self._cache = Cache(
                directory=self.cache_dir,
                size_limit=int(shutil.disk_usage(".").free * 0.95),
                cull_limit=0,
                tag_index=True,
            )

        logger.info(
            f"Connected cache '{self.cache_dir}' and export directory '{self.export_dir}'."
        )



import polars as pl
from typing import Any, Optional, Dict, Tuple, List


@pl.api.register_dataframe_namespace("tidy")
class TidyFrame:
    def __init__(self, df: pl.DataFrame):
        self._df = df

    def where(
        self,
        mapping: Optional[Dict[str, Tuple | List | Dict[int, str]]] = None,
    ) -> pl.DataFrame:
        pass


@pl.api.register_expr_namespace("tidy")
class TidyExpr:
    def __init__(self, expr: pl.Expr):
        self._expr = expr

    def trunc(self, range: Tuple[Any, Any] | List[Any]) -> pl.Expr:
        if isinstance(range, tuple):
            return (
                pl.when(self._expr.is_between(range[0], range[1], closed="both"))
                .then(self._expr)
                .otherwise(None)
            )
        else:
            return pl.when(self._expr.is_in(range)).then(self._expr).otherwise(None)

    def winsor(self, lower: Any, upper: Any) -> pl.Expr:
        return self._expr.clip(lower, upper)


@pl.api.register_series_namespace("tidy")
class TidySeries:
    def __init__(self, s: pl.Series):
        self._s = s

    def winsor(self) -> pl.Series:
        return self._s.cut()



from tidydata.base import TableBase
import pandas as pd
import polars as pl
from typing import Optional, List, Literal, Dict, Any, Tuple, Hashable, Iterable
from tidydata.dtype import DtypeStr, DtypeMappings
from tidydata.utils import (
    convert_dtype_backend,
    reshape_frame,
    concat_frames,
    log_error,
    split_frame,
)
from pathlib import Path
import numpy as np
import itertools
import warnings

pd.options.mode.copy_on_write = True
warnings.simplefilter(action="ignore", category=RuntimeWarning)


class Table(TableBase):
    def describe(
        self,
        destype: Literal["metafile", "summary"] = "summary",
    ):
        if destype == "metafile":
            described_df = pd.DataFrame(
                {
                    "Column": self.column_names,
                    "Column_Label": self.column_labels,
                    "Column_Alias": pd.Series(index=self.columns, dtype="string"),
                    "Column_Dtype": self.column_dtypes,
                    "Value_Range": self.column_ranges,
                    "Value_PreFill": pd.Series(index=self.columns, dtype="string"),
                    "Value_PreRepl": pd.Series(index=self.columns, dtype="string"),
                    "Is_Use": pd.Series(data=True, index=self.columns, dtype="bool"),
                }
            ).reset_index(drop=True)

        else:
            described_df = pd.DataFrame(
                {
                    "Column": self.column_names,
                    "Column_Label": self.column_labels,
                    "Column_Dtype": self.column_dtypes,
                    "Value_Range": self.column_ranges,
                    "Value_Missing": self.column_missings,
                }
            ).reset_index(drop=True)

        return self.__class__(
            data=described_df,
            name=f"{destype}_Name:{self.name}",
            tag=f"{destype}_Tag:{self.tag}",
            description=f"{destype}_{self.name}_Description:{self.description}",
            source=f"{destype}_{self.name}_Source:{self.source}",
        )

    def mutate(
        self,
        exprs: List[str],
        repl_dict: Optional[Dict[str, str]] = None,
        engine: Optional[Literal["numexpr", "python"]] = None,
    ):

        NA = pd.NA
        nan = np.nan
        NaT = pd.NaT
        empty = pd.Series(index=self.index,dtype='object')

        import tidydata.pandas_extension

        expr_df = (
            pd.Series(exprs)
            .str.strip()
            .str.split("=", n=1, expand=True)
            .rename(columns={0: "lhs", 1: "rhs"})
            .dropna(how="any")
        )

        if repl_dict is not None:
            expr_df["rhs"] = expr_df["rhs"].replace(repl_dict, "", regex=True)

        expr = expr_df["lhs"].str.cat(expr_df["rhs"], sep="=").str.cat(sep="\n")

        if engine is None:
            try:
                self.data = self.data.eval(expr, engine="numexpr")
            except:
                self.data = self.data.eval(expr, engine="python")
        else:
            self.data = self.data.eval(expr, engine=engine)

        return self

    def reshape(
        self,
        to_shape: Literal["long", "wide"],
        groups: str | List[str] | Dict[str, List[str]],
        i: Optional[str | List[str]] = None,
        j: Optional[str | List[str]] = None,
        sep: str = "_",
        suffix: str = r".*",
        dropna: bool = False
    ):  
        self.data = reshape_frame(
            self.data, to_shape, groups=groups, i=i, j=j, 
            sep=sep, suffix=suffix,dropna=dropna
        )
        
        

        return self

    def concat(
        self,
        others: Iterable["Table"],
        by_axis: Literal["column", "row"],
        on: Optional[List[str]] = None,
        how: Literal["left", "inner", "outer"] = "outer",
        keycol: Optional[Dict[str, int | str]] = None,
    ):

        
        self.data = concat_frames(
            itertools.chain([self.data], (oth.data for oth in others)),
            by_axis,
            on=on,
            how=how,
            keycol=keycol,
        )
        
        return self


    def format(
        self,
        name: Optional[str] = None,
        tag: Optional[str] = None,
        description: Optional[str] = None,
        columns: Optional[List[str] | str] = None,
        rows: Optional[Any] = None,
        column_labels: Optional[Dict[str, str]] = None,
        column_dtypes: Optional[Dict[str, DtypeStr]] = None,
        value_repls: Optional[Dict[str, Dict[Hashable, Any]]] = None,
        value_nafills: Optional[Dict[str, Any]] = None,
        value_ranges: Optional[Dict[str, Tuple | List | Dict[int, str]]] = None,
        dtype_backend: Optional[Literal["numpy", "numpy_nullable", "pyarrow"]] = None,
        drop_duplicates: Optional[List[str] | Literal["all"]] = None,
        column_aliases: Optional[Dict[str, str]] = None,
    ):
        return (
            self.reset_table_name(name)
            .reset_table_tag(tag)
            .reset_table_description(description)
            .select_columns(columns)
            .label_column_names(column_labels)
            .replace_column_values(value_repls)
            .fill_column_missings(value_nafills)
            .convert_column_dtypes(column_dtypes)
            .limit_column_ranges(value_ranges)
            .convert_dtype_backend(dtype_backend)
            .drop_duplicate_values(drop_duplicates)
            .select_rows(rows)
            .alias_column_names(column_aliases)
        )

    def split(
        self,
        by_columns: Optional[str | List[str]] = None,
        by_rows: Optional[int] = None,
    ):
        splited_dfs = split_frame(self.data, by_columns=by_columns, by_rows=by_rows)
        for k, df in splited_dfs:
            yield f"{self.name}_{k}", Table(
                data=df.reset_index(drop=True),
                name=f"{self.name}_{k}",
                tag=f"{self.tag}_{k}",
                description=self.description,
                source=self.source,
            )

    def to_datafile(self, path: Path | str, **options):
        from tidydata.io import write_file

        write_file(data=self, to_file=path, **options)

    def to_metafile(self, path: Path | str, **options):
        from tidydata.io import write_file

        write_file(data=self.describe(destype="metafile"), to_file=path, **options)

    def to_statfile(self, path: Path | str, **options):
        from tidydata.io import write_file

        write_file(data=self.describe(destype="summary"), to_file=path, **options)
